// Copyright (C) 1999-2000 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of ccscript.
// 
// The exception is that, if you link the ccscript library with other
// files to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the ccscript library code into it.
// 
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name ccscript.  If you copy code from other releases into a copy of
// ccscript, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for ccscript, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.  

#include <cc++/config.h>
#include <cc++/macros.h>
#include <cc++/misc.h>
#include <cc++/slog.h>
#include <stdlib.h>
#include <stdio.h>
#include "script.h"

static void logerror(const char *script, unsigned id, const char *msg)
{
	slog(SLOG_ERROR) << script;
	if(id)
		slog() << "(" << id << ")";
	slog() << ": " << msg << endl;
}

ScriptInterp::ScriptInterp(ScriptCommand *cmdset, int sym, int pg) :
ScriptSymbol(sym, pg)
{
	once = true;
	loop = true;
	signalmask = 0;
	packtoken = ',';
	stack = 0;
	cmd = cmdset;
	image = NULL;
	setSymbol("error", sym);
	setSymbol("error", "none");
};

void ScriptInterp::Advance(void)
{
	script[stack].line = script[stack].line->next;
	if(!script[stack].line && loop)
		script[stack].line = script[stack].script->first;
}

void ScriptInterp::Error(const char *errmsg)
{
	setSymbol("error", errmsg);
	Trap(1);
}

void ScriptInterp::Trap(const char *trapid)
{
	unsigned trap = cmd->getTrapId(trapid);
	if(!trap)
	{
		if(!stricmp(trapid, "first") || !stricmp(trapid, "top"))
		{
			script[stack].line = script[stack].script->first;
			return;
		}
	}
	Trap(trap);
}

void ScriptInterp::Trap(unsigned id)
{
	scriptline_t *trap = script[stack].script->trap[id];
	if(!trap && id)
	{
		Advance();
		return;
	}
	script[stack].line = trap;
}	

bool ScriptInterp::Step(const char *trapname)
{
	unsigned long mask;

	script[stack].index = 0;

	if(trapname)
		mask = cmd->getTrapMask(trapname);
	else
		mask = 0;

	if(!script[stack].line)
	{
		Exit();
		return false;
	}

	if(mask & script[stack].line->mask)
		Trap(trapname);

	return Execute(script[stack].line->method);
}

bool ScriptInterp::Signal(const char *trapname)
{
	unsigned long mask = cmd->getTrapMask(trapname);
	mask &= script[stack].line->mask;
	if(!mask)
		return false;

	Stop(mask);
	Trap(trapname);
	return true;
}

bool ScriptInterp::Signal(unsigned id)
{
	if(id >= TRAP_BITS)
		return false;

	unsigned long mask = cmd->getTrapMask(id);
	mask &= script[stack].line->mask;
	if(!mask)
	{
		signalmask |= id;
		return false;
	}

	Stop(mask);
	Trap(id);
	return true;
}

bool ScriptInterp::scrSlog(void)
{
	unsigned id = getId();
	char *val;
	slog(SLOG_INFO) << script[stack].script->name;
	if(id)
		slog() << "(" << id << ")";
	slog() << ": ";
	while(NULL != (val = getValue(NULL)))
		slog() << val;
	slog() << endl;
	Advance();
	return true;
}

bool ScriptInterp::scrElog(void)
{
	char *val;
	unsigned id = getId();

	slog(SLOG_ERROR) << script[stack].script->name;
	if(id)
		slog() << "(" << id << ")"; 
	slog() << ": ";
	while(NULL != (val = getValue(NULL)))
		slog() << val;
	slog() << endl;
	Advance();
	return true;
}

bool ScriptInterp::scrMissing(void)
{
	char *chk = getValue(NULL);

	if(!chk)
	{
		Error("missing-no-keyword");
		return true;
	}

	if(!cmd->getHandler(chk))
		return scrGoto();

	Advance();
	return true;
}

bool ScriptInterp::scrHas(void)
{
	char *chk = getValue(NULL);

	if(!chk)
	{
		Error("has-no-keyword");
		return true;
	}

	if(cmd->getHandler(chk))
		return scrGoto();

	Advance();
	return true;
}

bool ScriptInterp::scrSelect(void)
{
	char *value = getValue(NULL);
	char *match;

	if(!value)
	{
		Error("branch-no-select");
		return true;
	}

	while(NULL != (match = getValue(NULL)))
	{
		if(!stricmp(match, value))
			return scrGoto();

		getValue(NULL);
	}
	Advance();
	return true;
}		

bool ScriptInterp::scrOnce(void)
{
	if(getOnce())
		return scrGoto();
	Advance();
	return true;
}

bool ScriptInterp::scrOn(void)
{
	char *sig = getValue(NULL);
	unsigned long mask;

	if(!sig)
	{
		Error("on-no-signal");
		return true;
	}

	mask = cmd->getTrapMask(sig);
	if(!mask)
	{
		Error("on-invalid-signal");
		return true;
	}
	if(signalmask & mask)
	{
		signalmask &= ~mask;
		return scrGoto();
	}
	Advance();
	return true;
}

bool ScriptInterp::scrGoto(void)
{
	char namebuf[256];
	char *label = getValue(NULL);
	char *ext;
	scriptline_t *skip;
	scriptname_t *scr;

	if(!label)
	{
		Error("branch-failed");
		return true;
	}

	if(*label == '^')
	{
		Trap(++label);
		return true;
	}

	if(*label >= '0' && *label <= '9' && !label[1])
	{
		scriptline_t *skip = script[stack].script->skip[atoi(label)];
		if(!skip)
			Advance();

		return true;
	}

	if(!strncmp(label, "::", 2))
	{
		strcpy(namebuf, script[stack].script->name);
		ext = strstr(namebuf, "::");
		if(ext)
			strcpy(ext, label);		
		else
			strcat(namebuf, label);
		label = namebuf;
	}
	scr = image->getScript(label);
	if(!scr)
	{
		Error("script-not-found");
		return true;
	}
	once = true;
	script[stack].script = scr;
	script[stack].line = scr->first;
	return true;
}

bool ScriptInterp::scrFor(void)
{
	scriptsymbol_t *sym;
	unsigned short loop = script[stack].line->loop;
	scriptline_t *line;
	int index = script[stack].index;
	script[stack].index = 0;
	
	if(!index)
		++index;

	sym = getVariable(getSymbolSize());
	char *value;
	
	if(!sym)
	{
		Error("symbol-not-found");
		return true;
	}

	if(sym->flags.readonly)
	{
		Error("symbol-readonly");
		return true;
	}

	sym->data[sym->flags.size] = 0;
	script[stack].index = index;
	value = getValue(NULL);
	if(!value)
	{
		line = script[stack].line->next;
		while(line)
		{
			if(line->loop == loop)
			{
				script[stack].line = line;
				Advance();
				return true;
			}
			line = line->next;
		}
		Error("loop-overflow");
		return true;				
	}

	strncpy(sym->data, value, sym->flags.size);
	if(sym->flags.commit)
		Commit(sym);

	if(!Push())
		return true;

	Advance();
	return true;
}	
	
bool ScriptInterp::scrDo(void)
{
	unsigned short loop = script[stack].line->loop;
	scriptline_t *line;
		
	script[stack].index = 0;	// always reset

	if(script[stack].line->argc)
	{
		if(!Conditional())
		{
			line = script[stack].line->next;
			while(line)
			{
				if(line->loop == loop)
				{
					script[stack].line = line;
					Advance();
					return true;
				}
				line = line->next;
			}
			Error("loop-overflow");
			return true;
		}
	}

	if(!Push())
		return true;

	Advance();
	return true;
}
	
bool ScriptInterp::scrLoop(void)
{
	unsigned short loop;

	if(stack < 1)
	{
		Error("stack-underflow");
		return true;
	}

	loop = script[stack - 1].line->loop;
	if(!loop)
	{
		Error("stack-not-loop");
		return true;
	}
	
	if(script[stack].line->argc)
	{
		if(!Conditional())
		{
			script[stack - 1] = script[stack];
			--stack;
			Advance();
			return true;
		}
	}
	
	--stack;
	return Execute(script[stack].line->method);
}

bool ScriptInterp::scrContinue(void)
{
	scriptline_t *line;
	unsigned short loop;

	if(script[stack].line->argc)
	{
		if(!Conditional())
		{
			Advance();
			return true;
		}
	}

	if(stack < 1)
	{
		Error("stack-underflow");
		return true;
	}

	loop = script[stack - 1].line->loop;
	line = script[stack].line->next;
	
	if(!loop)
	{
		Error("stack-not-loop");
		return true;
	}

	while(line)
	{
		if(line->loop == loop)
		{
			script[stack].line = line;	
			return true;
		}
		line = line->next;
	}
	Error("loop-overflow");
	return true;
}

bool ScriptInterp::scrBreak(void)
{
	scriptline_t *line;
	unsigned short loop;

	if(script[stack].line->argc)
	{
		if(!Conditional())
		{
			Advance();
			return true;
		}
	}

	if(stack < 1)
	{
		Error("stack-underflow");
		return true;
	}
	
	loop = script[stack - 1].line->loop;
	line = script[stack].line->next;

	if(!loop)
	{
		Error("stack-not-loop");
		return true;
	}

	while(line)
	{
		if(line->loop == loop)
		{
			--stack;
			script[stack].line = line;
			Advance();
			return true;
		}
		line = line->next;
	}
	Error("loop-overflow");
	return true;
}

bool ScriptInterp::scrIf(void)
{
	if(Conditional())
		return scrGoto();

	Advance();
	return true;
}

bool ScriptInterp::scrTryLock(void)
{
	unsigned id = atoi(getValue("0"));
	
	if(!locks.Lock(this, id))
		Error("lock-busy");
	else
		Advance();

	return true;
}

bool ScriptInterp::scrWaitLock(void)
{
	unsigned id = atoi(getValue("0"));

	if(!locks.Lock(this, id))
		return true;

	Advance();
	return true;
}

bool ScriptInterp::scrUnlock(void)
{
	unsigned id = atoi(getValue("0"));

	if(!locks.Unlock(this, id))
		Error("lock-invalid");
	else
		Advance();

	return true;
}

bool ScriptInterp::scrCall(void)
{
	if(!Push())
		return true;

	return scrGoto(); 
}

bool ScriptInterp::scrPop(void)
{
	char *level = getValue(NULL);
	int max;

	if(level)
	{
		max = atoi(level);
		if(max < stack)
			stack = max;
		Advance();
		return true;
	}

	if(stack < 1)
	{
		Error("stack-underflow");
		return true;
	}
	script[stack - 1] = script[stack];
	--stack;
	Advance();
	return true;
}

bool ScriptInterp::scrReturn(void)
{
	if(!Pull())
		return true;

	Advance();
	return true;
}
		
bool ScriptInterp::scrExit(void)
{
	stack = 0;
	script[stack].line = NULL;
	return true;
}

bool ScriptInterp::scrClear(void)
{
	scriptsymbol_t *sym = getVariable(0);

	while(sym)
	{
		sym->data[0] = 0;
		sym->flags.initial = true;
		if(sym->flags.commit)
			Commit(sym);
		sym = getVariable(0);
	}
	Advance();
	return true;
}

bool ScriptInterp::scrSize(void)
{
	scriptsymbol_t *sym;
	int size = atoi(getValue("0"));
	char *opt = getOption(NULL);

	if(!opt)
	{
		Error("symbol-not-specified");
		return true;
	}

	if(!size)
	{
		Error("symbol-no-size");
		return true;
	}

	while(opt)
	{
		if(*opt == '@')
		{
			opt = getSymbol(++opt);
			if(!opt)
			{
				opt = getOption(NULL);
				continue;
			}
		}
		else if(*opt != '%')
		{
			Error("symbol-not-constant");
			return true;
		}
	
		if(*opt == '%')
			++opt;
		sym = getEntry(opt, size);
		opt = getOption(NULL);
	}

	Advance();
	return true;
}

bool ScriptInterp::scrInc(void)
{
	scriptsymbol_t *sym = getVariable(getSymbolSize());
	int len;
	int value;
	char *opt;

	if(!sym)
	{
		Error("symbol-not-found");
		return true;
	}

	if(sym->flags.readonly)
	{
		Error("symbol-readonly");
		return true;
	}

	value = atoi(sym->data);
	len = sym->flags.size;
	opt = getValue(NULL);
	if(opt)
	{
		while(opt)
		{
			value += atoi(opt);
			opt = getValue(NULL);
		}
	}
	else
		++value;

	sprintf(sym->data, "%d", value);
	if(sym->flags.commit)
		Commit(sym);

	Advance();
	return true;
}

bool ScriptInterp::scrDec(void)
{
	scriptsymbol_t *sym = getVariable(getSymbolSize());
	int len;
	int value;
	char *opt;

	if(!sym)
	{
		Error("symbol-not-found");
		return true;
	}

	if(sym->flags.readonly)
	{
		Error("symbol-readonly");
		return true;
	}

	value = atoi(sym->data);
	len = sym->flags.size;
	opt = getValue(NULL);
	if(opt)
	{
		while(opt)
		{
			value -= atoi(opt);
			opt = getValue(NULL);
		}
	}
	else
		--value;

	sprintf(sym->data, "%d", value);
	if(sym->flags.commit)
		Commit(sym);

	Advance();
	return true;
}

bool ScriptInterp::scrUnpack(void)
{
	scriptsymbol_t *sym;
	char *pp;
	char *name;
	char *gp = getValue(NULL);
	int len;

	if(!gp)
	{
		Error("source-not-found");
		return true;
	}

	while(*gp && NULL != (name = getOption(NULL)))
	{
		if(*name == '@')
			sym = getIndirect(++name);
		if(*name == '%')
			sym = getEntry(++name, getSymbolSize());
		else
			sym = NULL;

		if(sym)
		{
			if(sym->flags.readonly)
				pp = NULL;
			else
				pp = sym->data;
		}
		else
			pp = NULL;

		len = 0;
		while(pp && *gp && *gp != packtoken)
		{
			if(++len > sym->flags.size)
				break;

			*(pp++) = *(gp++);
		}

		if(pp)
		{
			*pp = 0;
			sym->flags.initial = false;
			if(sym->flags.commit)
				Commit(sym);
		}

		while(*gp && *gp != packtoken)
			++gp;

		if(*gp == packtoken)
			++gp;
	}
	Advance();
	return true;
}		

bool ScriptInterp::scrPack(void)
{
	scriptsymbol_t *sym;
	char *name = getOption(NULL);
	char *value;
	int len = 0;
	char *pp;

	if(!name)
	{
		Error("symbol-not-found");
		return true;
	}

	if(*name == '@')
		sym = getIndirect(++name);
	else if(*name == '%')
		sym = getEntry(++name, getSymbolSize());

	if(!sym)
	{
		Error("symbol-not-found");
		return true;
	}

	if(sym->flags.readonly)
	{
		Error("symbol-read-only");
		return true;
	}

	pp = sym->data;
	while(len < sym->flags.size && NULL != (value = getValue(NULL)))
	{
		while(*value && len++ < sym->flags.size)
			*(pp++) = *(value++);

		if(len < sym->flags.size)
			*(pp++) = packtoken;
	}
	sym->flags.initial = false;
	if(sym->flags.commit)
		Commit(sym);
	Advance();
	return true;
}

bool ScriptInterp::scrPacked(void)
{
	char *value = getValue(NULL);
	if(value)
		packtoken = *value;
	Advance();
	true;
}

bool ScriptInterp::scrSet(void)
{
	scriptsymbol_t *sym = getVariable(getSymbolSize());
	int len;
	char *value;

	if(!sym)
	{
		Error("symbol-not-found");
		return true;
	}

	if(sym->flags.readonly)
	{
		Error("symbol-readonly");
		return true;
	}

	sym->data[sym->flags.size] = 0;
	len = 0;

	while(len < sym->flags.size && NULL != (value = getValue(NULL)))
	{
		strncpy(sym->data + len, value, sym->flags.size - len);
		len = strlen(sym->data);	
	}

	sym->flags.initial = false;
	if(sym->flags.commit)
		Commit(sym);

	Advance();
	return true;
}

bool ScriptInterp::scrInit(void)
{
	scriptsymbol_t *sym = getVariable(getSymbolSize());
	int len;
	char *value;

	if(!sym)
	{
		Error("symbol-not-found");
		return true;
	}

	if(!sym->flags.initial)
	{
		Advance();
		return true;
	}

	if(sym->flags.readonly)
	{
		Error("symbol-readonly");
		return true;
	}

	sym->data[sym->flags.size] = 0;
	len = 0;

	while(len < sym->flags.size && NULL != (value = getValue(NULL)))
	{
		strncpy(sym->data + len, value, sym->flags.size - len);
		len = strlen(sym->data);	
	}

	sym->flags.initial = false;
	if(sym->flags.commit)
		Commit(sym);

	Advance();
	return true;
}

bool ScriptInterp::Conditional(void)
{
	char *v1 = getOption(NULL);
	char *op = getValue(NULL);
	char *v2 = getOption(NULL);
	int l1, l2;

	// no first parm, invalid

	if(!v1)
		return false;
	
	v1 = getContent(v1);

	// sym/label by itself, invalid

	if(!op)
	{
		script[stack].index = 0;
		return false;
	}
	
	// ifdef sym ... format assumed

	if(!v2)
	{
		script[stack].index = 1;
		if(*v1)
			return true;
		return false;
	}
	
	v2 = getContent(v2);

	if(!v1)
		v1 = "";

	if(!v2)
		v2 = "";
	
	if(!stricmp(op, "="))
	{
		if(atol(v1) == atol(v2))
			return true;
		return false;
	}

	if(!stricmp(op, "<>"))
	{
		if(atol(v1) != atol(v2))
			return true;
		return false;
	}

	if(!stricmp(op, "=="))
	{
		if(!stricmp(v1, v2))
			return true;

		return false;
	}

	if(!stricmp(op, "!="))
	{
		if(stricmp(v1, v2))
			return true;

		return false;
	}

	if(!stricmp(op, "$"))
	{
		if(strstr(v2, v1))
			return true;
		return false;
	}

	if(!stricmp(op, "$<") || !stricmp(op, "$+"))
	{
		if(!strnicmp(v1, v2, strlen(v1)))
			return true;
		return false;
	}

	if(!stricmp(op, "$>") || !stricmp(op, "$-"))
	{
		l1 = strlen(v1);
		l2 = strlen(v2);
		if(l1 <= l2)
			if(!strnicmp(v1, v2 + l2 - l1, l1))
				return true;

		return false;
	}

	if(!stricmp(op, "<"))
	{
		if(atol(v1) < atol(v2))
			return true;
		return false;
	}

	if(!stricmp(op, "<=") || !stricmp(op, "=<"))
	{
		if(atol(v1) <= atol(v2))
			return true;
		return false;
	}

	if(!stricmp(op, ">"))
	{
		if(atol(v1) > atol(v2))
			return true;
		return false;
	}

	if(!stricmp(op, ">=") || !stricmp(op, "=>"))
	{
		if(atol(v1) >= atol(v2))
			return true;
		return false;
	}

	// if no op, assume ifdef format

	script[stack].index = 1;
	if(*v1)
		return true;

	return false;
}		
		
bool ScriptInterp::Push(void)
{
	if(stack >= (SCRIPT_STACK_SIZE - 1))
	{
		Error("stack-overflow");
		return false;
	}

	script[stack + 1] = script[stack];
	++stack;
	return true;
}

bool ScriptInterp::Pull(void)
{
	if(!stack)
	{
		Error("stack-underflow");
		return false;
	}

	--stack;
	return true;
}

bool ScriptInterp::getOnce(void)
{
	bool ret = once;
	once = false;
	return ret;
}

bool ScriptInterp::Redirect(const char *scriptname)
{
	scriptname_t *scr;
	char namebuf[128];
	char *ext;

	if(!strncmp(scriptname, "::", 2))
	{
		strcpy(namebuf, script[stack].script->name);
		ext = strstr(namebuf, "::");
		if(ext) 
			strcpy(ext, scriptname);
		else
			strcat(namebuf, scriptname);
	}
	else
		strcpy(namebuf, scriptname);
	
	scr = image->getScript(namebuf);
	if(scr)
	{
		script[stack].script = scr;
		script[stack].line = scr->first;
		return true;
	}
	return false;
}

bool ScriptInterp::Attach(const char *scriptname)
{
	stack = 0;
	cmd->EnterMutex();
	image = cmd->active;
	if(!image)
	{
		cmd->LeaveMutex();
		return false;
	}

	script[stack].script = image->getScript(scriptname);
	if(script[stack].script)
	{
		script[stack].line = script[stack].script->first;
		++image->refcount;
		cmd->LeaveMutex();
		return true;
	}
	cmd->LeaveMutex();
	packtoken = ',';
	once = true;
	signalmask = 0;
	logerror(scriptname, getId(), "missing; attach failed");
	return false;	
}

void ScriptInterp::Detach(void)
{
	if(!image)
		return;

	cmd->EnterMutex();
	--image->refcount;
	if(!image->refcount && image != cmd->active)
		delete image;
	cmd->LeaveMutex();
	image = NULL;

	stack = 0;
	setSymbol("error", "none");
	locks.Release(this);
}

char *ScriptInterp::getOption(const char *def)
{
	if(script[stack].index >= script[stack].line->argc)
		return (char *)def;

	return script[stack].line->args[script[stack].index++];
}

scriptsymbol_t *ScriptInterp::getVariable(int size)
{
	char *opt;
	scriptsymbol_t *sym;

	if(script[stack].index >= script[stack].line->argc)
		return NULL;

	opt = script[stack].line->args[script[stack].index++];
	if(*opt != '%' && *opt != '@')
	{
		Error("variable-not-constant");
		return NULL;	
	}
	
	if(*opt == '@')
	{
		sym = getIndirect(++opt);
		if(sym)
			return sym;

		sym = getEntry(opt);
		if(sym)
			return getEntry(sym->data, size);
	}

	return getEntry(++opt, size);
}

char *ScriptInterp::getContent(char *opt)
{
	scriptsymbol_t *sym;

	if(!opt)
		return NULL;

	if(*opt != '%' && *opt != '@')
		return opt;

	if(*opt == '@')
	{
		++opt;
		sym = getIndirect(opt);
		if(sym)
			return sym->data;

		sym = getEntry(opt);
		if(!sym)
			return NULL;

		opt = sym->data;
	}
	else
		++opt;

	sym = getEntry(opt);
	if(!sym)
		return NULL;

	return sym->data;
}

char *ScriptInterp::getValue(const char *def)
{
	char *opt = getOption(NULL);

	if(!opt)
		return (char *)def;

	opt = getContent(opt);
	if(!opt)
		return (char *)def;

	return opt;
}	
